@@ -9,12 +9,17 @@ import yaml
 
 from ._utils import (
     ValueKind,
+    _ensure_container,
     _get_value,
     _is_interpolation,
+    _resolve_optional,
+    get_ref_type,
     get_value_kind,
     get_yaml_loader,
-    is_primitive_container,
+    is_dict_annotation,
+    is_list_annotation,
     is_primitive_dict,
+    is_primitive_type,
     is_structured_config,
 )
 from .base import Container, ContainerMetadata, Node
@@ -28,10 +33,13 @@ class BaseContainer(Container, ABC):
     _resolvers: Dict[str, Any] = {}
 
     def __init__(self, parent: Optional["Container"], metadata: ContainerMetadata):
-        super().__init__(
-            parent=parent, metadata=metadata,
-        )
+        super().__init__(parent=parent, metadata=metadata)
         self.__dict__["_content"] = None
+        self._normalize_ref_type()
+
+    def _normalize_ref_type(self) -> None:
+        if self._metadata.ref_type is None:
+            self._metadata.ref_type = Any  # type: ignore
 
     def _resolve_with_default(
         self,
@@ -211,59 +219,73 @@ class BaseContainer(Container, ABC):
         assert False
 
     def pretty(self, resolve: bool = False, sort_keys: bool = False) -> str:
-        """
-        returns a yaml dump of this config object.
-        :param resolve: if True, will return a string with the interpolations resolved, otherwise
-        interpolations are preserved
-        :param sort_keys: If True, will print dict keys in sorted order. default False.
-        :return: A string containing the yaml representation.
-        """
         from omegaconf import OmegaConf
 
-        container = OmegaConf.to_container(self, resolve=resolve, enum_to_str=True)
-        return yaml.dump(  # type: ignore
-            container, default_flow_style=False, allow_unicode=True, sort_keys=sort_keys
+        warnings.warn(
+            """
+            pretty() is deprecated and will be removed in a future version.
+            Use OmegaConf.to_yaml. Please note that the default value for
+            resolve has changed to True.
+            """,
+            category=UserWarning,
         )
 
+        return OmegaConf.to_yaml(self, resolve=resolve, sort_keys=sort_keys)
+
     @staticmethod
     def _map_merge(dest: "BaseContainer", src: "BaseContainer") -> None:
         """merge src into dest and return a new copy, does not modified input"""
-        from omegaconf import OmegaConf
-
-        from .dictconfig import DictConfig
-        from .nodes import ValueNode
+        from omegaconf import DictConfig, OmegaConf, ValueNode
 
         assert isinstance(dest, DictConfig)
         assert isinstance(src, DictConfig)
         src_type = src._metadata.object_type
 
+        # if source DictConfig is missing set the DictConfig one to be missing too.
+        if src._is_missing():
+            dest._set_value("???")
+            return
         dest._validate_set_merge_impl(key=None, value=src, is_assign=False)
+
+        def expand(node: Container) -> None:
+            type_ = get_ref_type(node)
+            if type_ is not None:
+                _is_optional, type_ = _resolve_optional(type_)
+                if is_dict_annotation(type_):
+                    node._set_value({})
+                elif is_list_annotation(type_):
+                    node._set_value([])
+                else:
+                    node._set_value(type_)
+
+        if dest._is_missing():
+            expand(dest)
+
         for key, src_value in src.items_ex(resolve=False):
-            dest_element_type = dest._metadata.element_type
-            element_typed = dest_element_type not in (None, Any)
-            if OmegaConf.is_missing(dest, key):
-                if isinstance(src_value, DictConfig):
-                    if OmegaConf.is_missing(dest, key):
-                        dest[key] = src_value
 
             dest_node = dest._get_node(key, validate_access=False)
-            if dest_node is not None and dest_node._is_interpolation():
-                target_node = dest_node._dereference_node(
-                    throw_on_resolution_failure=False
-                )
-                if isinstance(target_node, Container):
-                    dest[key] = target_node
-                    dest_node = dest._get_node(key)
+            if isinstance(dest_node, Container) and OmegaConf.is_none(dest, key):
+                if not OmegaConf.is_none(src_value):
+                    expand(dest_node)
 
-            if dest_node is not None or element_typed:
-                if dest_node is None and element_typed:
-                    dest[key] = DictConfig(content=dest_element_type, parent=dest)
-                    dest_node = dest._get_node(key)
+            if dest_node is not None:
+                if dest_node._is_interpolation():
+                    target_node = dest_node._dereference_node(
+                        throw_on_resolution_failure=False
+                    )
+                    if isinstance(target_node, Container):
+                        dest[key] = target_node
+                        dest_node = dest._get_node(key)
 
+            if is_structured_config(dest._metadata.element_type):
+                dest[key] = DictConfig(content=dest._metadata.element_type, parent=dest)
+                dest_node = dest._get_node(key)
+
+            if dest_node is not None:
                 if isinstance(dest_node, BaseContainer):
                     if isinstance(src_value, BaseContainer):
                         dest._validate_merge(key=key, value=src_value)
-                        dest_node.merge_with(src_value)
+                        dest_node._merge_with(src_value)
                     else:
                         dest.__setitem__(key, src_value)
                 else:
@@ -273,60 +295,90 @@ class BaseContainer(Container, ABC):
                         assert isinstance(dest_node, ValueNode)
                         try:
                             dest_node._set_value(src_value)
-                        except ValidationError as e:
+                        except (ValidationError, ReadonlyConfigError) as e:
                             dest._format_and_raise(key=key, value=src_value, cause=e)
             else:
-                dest[key] = src._get_node(key)
+                from omegaconf import open_dict
+
+                if is_structured_config(src_type):
+                    # verified to be compatible above in _validate_set_merge_impl
+                    with open_dict(dest):
+                        dest[key] = src._get_node(key)
+                else:
+                    dest[key] = src._get_node(key)
 
         if src_type is not None and not is_primitive_dict(src_type):
             dest._metadata.object_type = src_type
 
+        # explicit flags on the source config are replacing the flag values in the destination
+        for flag, value in src._metadata.flags.items():
+            if value is not None:
+                dest._set_flag(flag, value)
+
     def merge_with(
         self,
         *others: Union["BaseContainer", Dict[str, Any], List[Any], Tuple[Any], Any],
+    ) -> None:
+        try:
+            self._merge_with(*others)
+        except Exception as e:
+            self._format_and_raise(key=None, value=None, cause=e)
+
+    def _merge_with(
+        self,
+        *others: Union["BaseContainer", Dict[str, Any], List[Any], Tuple[Any], Any],
     ) -> None:
         from .dictconfig import DictConfig
         from .listconfig import ListConfig
         from .omegaconf import OmegaConf
 
-        try:
-            """merge a list of other Config objects into this one, overriding as needed"""
-            for other in others:
-                if other is None:
-                    raise ValueError("Cannot merge with a None config")
-                if is_primitive_container(other):
-                    assert isinstance(other, (list, dict))
-                    other = OmegaConf.create(other)
-                elif is_structured_config(other):
-                    other = OmegaConf.structured(other)
-                if isinstance(self, DictConfig) and isinstance(other, DictConfig):
-                    BaseContainer._map_merge(self, other)
-                elif isinstance(self, ListConfig) and isinstance(other, ListConfig):
-                    if self._get_flag("readonly"):
-                        raise ReadonlyConfigError(self._get_full_key(""))
-                    if (
-                        self._is_none()
-                        or self._is_missing()
-                        or self._is_interpolation()
-                    ):
-                        self.__dict__["_content"] = []
-                    else:
-                        self.__dict__["_content"].clear()
-                    for item in other:
-                        self.append(item)
+        """merge a list of other Config objects into this one, overriding as needed"""
+        for other in others:
+            if other is None:
+                raise ValueError("Cannot merge with a None config")
+            other = _ensure_container(other)
+            if isinstance(self, DictConfig) and isinstance(other, DictConfig):
+                BaseContainer._map_merge(self, other)
+            elif isinstance(self, ListConfig) and isinstance(other, ListConfig):
+                if self._is_none() or self._is_missing() or self._is_interpolation():
+                    self.__dict__["_content"] = []
                 else:
-                    raise TypeError("Cannot merge DictConfig with ListConfig")
+                    self.__dict__["_content"].clear()
 
-            # recursively correct the parent hierarchy after the merge
-            self._re_parent()
-        except Exception as e:
-            self._format_and_raise(key=None, value=None, cause=e)
+                if other._is_missing():
+                    self._set_value("???")
+                else:
+                    et = self._metadata.element_type
+                    if is_structured_config(et):
+                        prototype = OmegaConf.structured(et)
+                        for item in other:
+                            if isinstance(item, DictConfig):
+                                item = OmegaConf.merge(prototype, item)
+                            self.append(item)
+
+                    else:
+                        for item in other:
+                            self.append(item)
+
+                # explicit flags on the source config are replacing the flag values in the destination
+                for flag, value in other._metadata.flags.items():
+                    if value is not None:
+                        self._set_flag(flag, value)
+            else:
+                raise TypeError("Cannot merge DictConfig with ListConfig")
+
+        # recursively correct the parent hierarchy after the merge
+        self._re_parent()
 
     # noinspection PyProtectedMember
     def _set_item_impl(self, key: Any, value: Any) -> None:
+        """
+        Changes the value of the node key with the desired value. If the node key doesn't
+        exist it creates a new one.
+        """
         from omegaconf.omegaconf import OmegaConf, _maybe_wrap
 
-        from .nodes import ValueNode
+        from .nodes import AnyNode, ValueNode
 
         if isinstance(value, Node):
             try:
@@ -338,14 +390,12 @@ class BaseContainer(Container, ABC):
         else:
             self._validate_set(key, value)
 
+        if self._get_flag("readonly"):
+            raise ReadonlyConfigError("Cannot change read-only config container")
+
         input_config = isinstance(value, Container)
         target_node_ref = self._get_node(key)
         special_value = value is None or value == "???"
-        should_set_value = (
-            target_node_ref is not None
-            and isinstance(target_node_ref, Container)
-            and special_value
-        )
 
         input_node = isinstance(value, ValueNode)
         if isinstance(self.__dict__["_content"], dict):
@@ -355,6 +405,21 @@ class BaseContainer(Container, ABC):
 
         elif isinstance(self.__dict__["_content"], list):
             target_node = isinstance(target_node_ref, ValueNode)
+        # We use set_value if:
+        # 1. Target node is a container and the value is MISSING or None
+        # 2. Target node is a container and has an explicit ref_type
+        # 3. If the target is a NodeValue then it should set his value.
+        #  Furthermore if it's an AnyNode it should wrap when the input is
+        # a container and set when the input is an compatible type(primitive type).
+
+        should_set_value = target_node_ref is not None and (
+            (
+                isinstance(target_node_ref, Container)
+                and (special_value or target_node_ref._has_ref_type())
+            )
+            or (target_node and not isinstance(target_node_ref, AnyNode))
+            or (isinstance(target_node_ref, AnyNode) and is_primitive_type(value))
+        )
 
         def wrap(key: Any, val: Any) -> Node:
             is_optional = True
@@ -382,17 +447,17 @@ class BaseContainer(Container, ABC):
             v._set_key(value_key)
             self.__dict__["_content"][value_key] = v
 
-        if is_primitive_container(value):
-            self.__dict__["_content"][key] = wrap(key, value)
-        elif input_node and target_node:
+        if input_node and target_node:
             # both nodes, replace existing node with new one
             assign(key, value)
         elif not input_node and target_node:
             # input is not node, can be primitive or config
-            if input_config:
+            if should_set_value:
+                self.__dict__["_content"][key]._set_value(value)
+            elif input_config:
                 assign(key, value)
             else:
-                self.__dict__["_content"][key]._set_value(value)
+                self.__dict__["_content"][key] = wrap(key, value)
         elif input_node and not target_node:
             # target must be config, replace target with input node
             assign(key, value)
@@ -406,7 +471,7 @@ class BaseContainer(Container, ABC):
 
     @staticmethod
     def _item_eq(
-        c1: Container, k1: Union[str, int], c2: Container, k2: Union[str, int],
+        c1: Container, k1: Union[str, int], c2: Container, k2: Union[str, int]
     ) -> bool:
         v1 = c1._get_node(k1)
         v2 = c2._get_node(k2)
@@ -482,9 +547,7 @@ class BaseContainer(Container, ABC):
     def _value(self) -> Any:
         return self.__dict__["_content"]
 
-    def _get_full_key(
-        self, key: Union[str, Enum, int, slice, None], disable_warning: bool = False
-    ) -> str:
+    def _get_full_key(self, key: Union[str, Enum, int, slice, None]) -> str:
         from .listconfig import ListConfig
         from .omegaconf import _select_one
 
@@ -524,11 +587,7 @@ class BaseContainer(Container, ABC):
         if key is not None and key != "":
             assert isinstance(self, Container)
             cur, _ = _select_one(
-                c=self,
-                key=str(key),
-                throw_on_missing=False,
-                throw_on_type_error=False,
-                disable_warning=disable_warning,
+                c=self, key=str(key), throw_on_missing=False, throw_on_type_error=False
             )
             if cur is None:
                 cur = self
